@@ -36,16 +36,15 @@
 #include <osmocom/gsm/protocol/gsm_04_11.h>
 #include <osmocom/gsm/protocol/smpp34_osmocom.h>
 
-#include <openbsc/gsm_subscriber.h>
-#include <openbsc/debug.h>
-#include <openbsc/db.h>
-#include <openbsc/gsm_04_11.h>
-#include <openbsc/gsm_data.h>
-#include <openbsc/signal.h>
-#include <openbsc/transaction.h>
-#include <openbsc/gsm_subscriber.h>
-#include <openbsc/chan_alloc.h>
-#include <openbsc/vlr.h>
+#include <osmocom/msc/gsm_subscriber.h>
+#include <osmocom/msc/debug.h>
+#include <osmocom/msc/db.h>
+#include <osmocom/msc/gsm_04_11.h>
+#include <osmocom/msc/gsm_data.h>
+#include <osmocom/msc/signal.h>
+#include <osmocom/msc/transaction.h>
+#include <osmocom/msc/gsm_subscriber.h>
+#include <osmocom/msc/vlr.h>
 
 #include "smpp_smsc.h"
 
@@ -73,80 +72,103 @@ static struct vlr_subscr *subscr_by_dst(struct gsm_network *net,
 	return vsub;
 }
 
-/*! \brief find a TLV with given tag in list of libsmpp34 TLVs */
-static struct tlv_t *find_tlv(struct tlv_t *head, uint16_t tag)
+static int smpp34_submit_tlv_msg_payload(const struct tlv_t *t,
+					 const struct submit_sm_t *submit,
+					 const uint8_t **sms_msg,
+					 unsigned int *sms_msg_len)
 {
-	struct tlv_t *t;
-
-	for (t = head; t != NULL; t = t->next) {
-		if (t->tag == tag)
-			return t;
+	if (submit->sm_length) {
+		LOGP(DLSMS, LOGL_ERROR,
+		     "SMPP cannot have payload in TLV _and_ in the header\n");
+		return -1;
 	}
-	return NULL;
+	*sms_msg = t->value.octet;
+	*sms_msg_len = t->length;
+
+	return 0;
 }
 
 /*! \brief convert from submit_sm_t to gsm_sms */
 static int submit_to_sms(struct gsm_sms **psms, struct gsm_network *net,
 			 const struct submit_sm_t *submit)
 {
+	const uint8_t *sms_msg = NULL;
+	unsigned int sms_msg_len = 0;
 	struct vlr_subscr *dest;
+	uint16_t msg_ref = 0;
 	struct gsm_sms *sms;
 	struct tlv_t *t;
-	const uint8_t *sms_msg;
-	unsigned int sms_msg_len;
 	int mode;
+	int can_store_sms = ((submit->esm_class & SMPP34_MSG_MODE_MASK) != 2); /* != forward mode */
 
 	dest = subscr_by_dst(net, submit->dest_addr_npi,
 			     submit->dest_addr_ton,
 			     (const char *)submit->destination_addr);
-	if (!dest) {
+	if (!dest && !can_store_sms) {
 		LOGP(DLSMS, LOGL_NOTICE, "SMPP SUBMIT-SM for unknown subscriber: "
 		     "%s (NPI=%u)\n", submit->destination_addr,
 		     submit->dest_addr_npi);
 		return ESME_RINVDSTADR;
 	}
 
-	t = find_tlv(submit->tlv, TLVID_message_payload);
-	if (t) {
-		if (submit->sm_length) {
-			/* ERROR: we cannot have both! */
-			LOGP(DLSMS, LOGL_ERROR, "SMPP Cannot have payload in "
-				"TLV _and_ in the header\n");
+	smpp34_tlv_for_each(t, submit->tlv) {
+		switch (t->tag) {
+		case TLVID_message_payload:
+			if (smpp34_submit_tlv_msg_payload(t, submit, &sms_msg,
+							  &sms_msg_len) < 0) {
+				if (dest)
+					vlr_subscr_put(dest);
+				return ESME_ROPTPARNOTALLWD;
+			}
+			break;
+		case TLVID_user_message_reference:
+			msg_ref = t->value.val16;
+			break;
+		default:
+			break;
+		}
+	}
+
+	if (!sms_msg) {
+		if (submit->sm_length > 0 && submit->sm_length < 255) {
+			sms_msg = submit->short_message;
+			sms_msg_len = submit->sm_length;
+		} else {
+			LOGP(DLSMS, LOGL_ERROR,
+			     "SMPP neither message payload nor valid sm_length.\n");
 			vlr_subscr_put(dest);
-			return ESME_ROPTPARNOTALLWD;
+			return ESME_RINVPARLEN;
 		}
-		sms_msg = t->value.octet;
-		sms_msg_len = t->length;
-	} else if (submit->sm_length > 0 && submit->sm_length < 255) {
-		sms_msg = submit->short_message;
-		sms_msg_len = submit->sm_length;
-	} else {
-		LOGP(DLSMS, LOGL_ERROR,
-			"SMPP neither message payload nor valid sm_length.\n");
-		vlr_subscr_put(dest);
-		return ESME_RINVPARLEN;
 	}
 
 	sms = sms_alloc();
 	sms->source = SMS_SOURCE_SMPP;
 	sms->smpp.sequence_nr = submit->sequence_number;
+	sms->status_rep_req = submit->registered_delivery;
+	sms->msg_ref = msg_ref;
 
 	/* fill in the destination address */
 	sms->receiver = dest;
 	sms->dst.ton = submit->dest_addr_ton;
 	sms->dst.npi = submit->dest_addr_npi;
-	osmo_strlcpy(sms->dst.addr, dest->msisdn, sizeof(sms->dst.addr));
+	if (dest)
+		OSMO_STRLCPY_ARRAY(sms->dst.addr, dest->msisdn);
+	else
+		OSMO_STRLCPY_ARRAY(sms->dst.addr,
+				   (const char *)submit->destination_addr);
 
 	/* fill in the source address */
 	sms->src.ton = submit->source_addr_ton;
 	sms->src.npi = submit->source_addr_npi;
-	osmo_strlcpy(sms->src.addr, (char *)submit->source_addr,
-		     sizeof(sms->src.addr));
+	OSMO_STRLCPY_ARRAY(sms->src.addr, (char *)submit->source_addr);
 
-	if (submit->esm_class & 0x40)
+	if (submit->esm_class == SMPP34_DELIVERY_ACK)
+		sms->is_report = true;
+
+	if (submit->esm_class & SMPP34_UDHI_IND)
 		sms->ud_hdr_ind = 1;
 
-	if (submit->esm_class & 0x80) {
+	if (submit->esm_class & SMPP34_REPLY_PATH) {
 		sms->reply_path_req = 1;
 #warning Implement reply path
 	}
@@ -222,7 +244,7 @@ int handle_smpp_submit(struct osmo_esme *esme, struct submit_sm_t *submit,
 	sms->smpp.esme = esme;
 	sms->protocol_id = submit->protocol_id;
 
-	switch (submit->esm_class & 3) {
+	switch (submit->esm_class & SMPP34_MSG_MODE_MASK) {
 	case 0: /* default */
 	case 1: /* datagram */
 	case 3: /* store-and-forward */
@@ -419,7 +441,7 @@ void append_tlv_u16(tlv_t **req_tlv, uint16_t tag, uint16_t val)
 	memset(&tlv, 0, sizeof(tlv));
 	tlv.tag = tag;
 	tlv.length = 2;
-	tlv.value.val16 = htons(val);
+	tlv.value.val16 = val;
 	build_tlv(req_tlv, &tlv);
 }
 
@@ -501,7 +523,6 @@ static void smpp_cmd_free(struct osmo_smpp_cmd *cmd)
 	osmo_timer_del(&cmd->response_timer);
 	llist_del(&cmd->list);
 	vlr_subscr_put(cmd->vsub);
-	sms_free(cmd->sms);
 	talloc_free(cmd);
 }
 
@@ -518,21 +539,24 @@ void smpp_cmd_ack(struct osmo_smpp_cmd *cmd)
 	struct gsm_subscriber_connection *conn;
 	struct gsm_trans *trans;
 
+	if (cmd->is_report)
+		goto out;
+
 	conn = connection_for_subscr(cmd->vsub);
 	if (!conn) {
 		LOGP(DSMPP, LOGL_ERROR, "No connection to subscriber anymore\n");
-		return;
+		goto out;
 	}
 
-	trans = trans_find_by_id(conn, GSM48_PDISC_SMS,
-				 cmd->sms->gsm411.transaction_id);
+	trans = trans_find_by_id(conn, GSM48_PDISC_SMS, cmd->gsm411_trans_id);
 	if (!trans) {
 		LOGP(DSMPP, LOGL_ERROR, "GSM transaction %u is gone\n",
-		     cmd->sms->gsm411.transaction_id);
-		return;
+		     cmd->gsm411_trans_id);
+		goto out;
 	}
 
-	gsm411_send_rp_ack(trans, cmd->sms->gsm411.msg_ref);
+	gsm411_send_rp_ack(trans, cmd->gsm411_msg_ref);
+out:
 	smpp_cmd_free(cmd);
 }
 
@@ -542,25 +566,27 @@ void smpp_cmd_err(struct osmo_smpp_cmd *cmd, uint32_t status)
 	struct gsm_trans *trans;
 	int gsm411_cause;
 
+	if (cmd->is_report)
+		goto out;
+
 	conn = connection_for_subscr(cmd->vsub);
 	if (!conn) {
 		LOGP(DSMPP, LOGL_ERROR, "No connection to subscriber anymore\n");
-		return;
+		goto out;
 	}
 
-	trans = trans_find_by_id(conn, GSM48_PDISC_SMS,
-				 cmd->sms->gsm411.transaction_id);
+	trans = trans_find_by_id(conn, GSM48_PDISC_SMS, cmd->gsm411_trans_id);
 	if (!trans) {
 		LOGP(DSMPP, LOGL_ERROR, "GSM transaction %u is gone\n",
-		     cmd->sms->gsm411.transaction_id);
-		return;
+		     cmd->gsm411_trans_id);
+		goto out;
 	}
 
 	if (smpp_to_gsm411_err(status, &gsm411_cause) < 0)
 		gsm411_cause = GSM411_RP_CAUSE_MO_NET_OUT_OF_ORDER;
 
-	gsm411_send_rp_error(trans, cmd->sms->gsm411.msg_ref, gsm411_cause);
-
+	gsm411_send_rp_error(trans, cmd->gsm411_msg_ref, gsm411_cause);
+out:
 	smpp_cmd_free(cmd);
 }
 
@@ -571,7 +597,7 @@ static void smpp_deliver_sm_cb(void *data)
 
 static int smpp_cmd_enqueue(struct osmo_esme *esme,
 			    struct vlr_subscr *vsub, struct gsm_sms *sms,
-			    uint32_t sequence_number, bool *deferred)
+			    uint32_t sequence_number)
 {
 	struct osmo_smpp_cmd *cmd;
 
@@ -580,7 +606,9 @@ static int smpp_cmd_enqueue(struct osmo_esme *esme,
 		return -1;
 
 	cmd->sequence_nr	= sequence_number;
-	cmd->sms		= sms;
+	cmd->is_report		= sms->is_report;
+	cmd->gsm411_msg_ref	= sms->gsm411.msg_ref;
+	cmd->gsm411_trans_id	= sms->gsm411.transaction_id;
 	cmd->vsub		= vlr_subscr_get(vsub);
 
 	/* FIXME: No predefined value for this response_timer as specified by
@@ -591,7 +619,6 @@ static int smpp_cmd_enqueue(struct osmo_esme *esme,
 	osmo_timer_setup(&cmd->response_timer, smpp_deliver_sm_cb, cmd);
 	osmo_timer_schedule(&cmd->response_timer, 5, 0);
 	llist_add_tail(&cmd->list, &esme->smpp_cmd_list);
-	*deferred = true;
 
 	return 0;
 }
@@ -609,8 +636,7 @@ struct osmo_smpp_cmd *smpp_cmd_find_by_seqnum(struct osmo_esme *esme,
 }
 
 static int deliver_to_esme(struct osmo_esme *esme, struct gsm_sms *sms,
-			   struct gsm_subscriber_connection *conn,
-			   bool *deferred)
+			   struct gsm_subscriber_connection *conn)
 {
 	struct deliver_sm_t deliver;
 	int mode, ret;
@@ -641,15 +667,20 @@ static int deliver_to_esme(struct osmo_esme *esme, struct gsm_sms *sms,
 	memcpy(deliver.destination_addr, sms->dst.addr,
 		sizeof(deliver.destination_addr));
 
-	deliver.esm_class	= 1;	/* datagram mode */
+	if (sms->is_report)
+		deliver.esm_class = SMPP34_DELIVERY_RECEIPT;
+	else
+		deliver.esm_class = SMPP34_DATAGRAM_MODE;
+
 	if (sms->ud_hdr_ind)
-		deliver.esm_class |= 0x40;
+		deliver.esm_class |= SMPP34_UDHI_IND;
 	if (sms->reply_path_req)
-		deliver.esm_class |= 0x80;
+		deliver.esm_class |= SMPP34_REPLY_PATH;
 
 	deliver.protocol_id 	= sms->protocol_id;
 	deliver.priority_flag	= 0;
-	deliver.registered_delivery = 0;
+	if (sms->status_rep_req)
+		deliver.registered_delivery = SMPP34_DELIVERY_RECEIPT_ON;
 
 	/* Figure out SMPP DCS from TP-DCS */
 	dcs = sms->data_coding_scheme;
@@ -679,8 +710,6 @@ static int deliver_to_esme(struct osmo_esme *esme, struct gsm_sms *sms,
 	} else {
 		deliver.sm_length = sms->user_data_len;
 		memcpy(deliver.short_message, sms->user_data, deliver.sm_length);
-		deliver.sm_length = sms->user_data_len;
-		memcpy(deliver.short_message, sms->user_data, deliver.sm_length);
 	}
 
 #if BEFORE_MSCSPLIT
@@ -689,12 +718,15 @@ static int deliver_to_esme(struct osmo_esme *esme, struct gsm_sms *sms,
 		append_osmo_tlvs(&deliver.tlv, conn->lchan);
 #endif
 
+	append_tlv_u16(&deliver.tlv, TLVID_user_message_reference,
+		       sms->msg_ref);
+
 	ret = smpp_tx_deliver(esme, &deliver);
 	if (ret < 0)
 		return ret;
 
 	return smpp_cmd_enqueue(esme, conn->vsub, sms,
-				deliver.sequence_number, deferred);
+				deliver.sequence_number);
 }
 
 static struct smsc *g_smsc;
@@ -705,21 +737,22 @@ int smpp_route_smpp_first(struct gsm_sms *sms, struct gsm_subscriber_connection
 }
 
 int smpp_try_deliver(struct gsm_sms *sms,
-		     struct gsm_subscriber_connection *conn, bool *deferred)
+		     struct gsm_subscriber_connection *conn)
 {
 	struct osmo_esme *esme;
 	struct osmo_smpp_addr dst;
+	int rc;
 
 	memset(&dst, 0, sizeof(dst));
 	dst.ton = sms->dst.ton;
 	dst.npi = sms->dst.npi;
 	memcpy(dst.addr, sms->dst.addr, sizeof(dst.addr));
 
-	esme = smpp_route(g_smsc, &dst);
-	if (!esme)
-		return GSM411_RP_CAUSE_MO_NUM_UNASSIGNED;
+	rc = smpp_route(g_smsc, &dst, &esme);
+	if (!rc)
+		rc = deliver_to_esme(esme, sms, conn);
 
-	return deliver_to_esme(esme, sms, conn, deferred);
+	return rc;
 }
 
 struct smsc *smsc_from_vty(struct vty *v)
