@@ -41,8 +41,8 @@
 
 #include "smpp_smsc.h"
 
-#include <openbsc/debug.h>
-#include <openbsc/gsm_data.h>
+#include <osmocom/msc/debug.h>
+#include <osmocom/msc/gsm_data.h>
 
 /*! \brief Ugly wrapper. libsmpp34 should do this itself! */
 #define SMPP34_UNPACK(rc, type, str, data, len)		\
@@ -249,6 +249,8 @@ static void esme_destroy(struct osmo_esme *esme)
 	}
 	smpp_cmd_flush_pending(esme);
 	llist_del(&esme->list);
+	if (esme->acl)
+		esme->acl->esme = NULL;
 	talloc_free(esme);
 }
 
@@ -270,8 +272,7 @@ void smpp_esme_put(struct osmo_esme *esme)
 }
 
 /*! \brief try to find a SMPP route (ESME) for given destination */
-struct osmo_esme *
-smpp_route(const struct smsc *smsc, const struct osmo_smpp_addr *dest)
+int smpp_route(const struct smsc *smsc, const struct osmo_smpp_addr *dest, struct osmo_esme **pesme)
 {
 	struct osmo_smpp_route *r;
 	struct osmo_smpp_acl *acl = NULL;
@@ -314,15 +315,20 @@ smpp_route(const struct smsc *smsc, const struct osmo_smpp_addr *dest)
 		struct osmo_esme *esme;
 		DEBUGP(DSMPP, "ACL even has ESME, we can route to it!\n");
 		esme = acl->esme;
-		if (esme->bind_flags & ESME_BIND_RX)
-			return esme;
-		else
+		if (esme->bind_flags & ESME_BIND_RX) {
+			*pesme = esme;
+			return 0;
+		} else
 			LOGP(DSMPP, LOGL_NOTICE, "[%s] is matching route, "
 			     "but not bound for Rx, discarding MO SMS\n",
 				     esme->system_id);
 	}
 
-	return NULL;
+	*pesme = NULL;
+	if (acl)
+		return GSM48_CC_CAUSE_NETWORK_OOO;
+	else
+		return GSM48_CC_CAUSE_UNASSIGNED_NR;
 }
 
 
@@ -505,8 +511,7 @@ static int smpp_handle_bind_tx(struct osmo_esme *esme, struct msgb *msg)
 	bind_r.command_status = rc;
 
 	/* build response */
-	snprintf((char *)bind_r.system_id, sizeof(bind_r.system_id), "%s",
-		 esme->smsc->system_id);
+	osmo_strlcpy((char*)bind_r.system_id, esme->smsc->system_id, sizeof(bind_r.system_id));
 
 	/* add interface version TLV */
 	tlv.tag = TLVID_sc_interface_version;
@@ -654,6 +659,9 @@ int smpp_tx_deliver(struct osmo_esme *esme, struct deliver_sm_t *deliver)
 {
 	deliver->sequence_number = esme_inc_seq_nr(esme);
 
+	LOGP(DSMPP, LOGL_DEBUG, "[%s] Tx DELIVER-SM (from %s)\n",
+		esme->system_id, deliver->source_addr);
+
 	return PACK_AND_SEND(esme, deliver);
 }
 
@@ -736,7 +744,7 @@ static int smpp_pdu_rx(struct osmo_esme *esme, struct msgb *msg __uses)
 	int rc = 0;
 
 	LOGP(DSMPP, LOGL_DEBUG, "[%s] smpp_pdu_rx(%s)\n", esme->system_id,
-	     osmo_hexdump(msgb_data(msg), msgb_length(msg)));
+	     msgb_hexdump(msg));
 
 	switch (cmd_id) {
 	case GENERIC_NACK:
@@ -869,6 +877,8 @@ dead_socket:
 	osmo_fd_unregister(&esme->wqueue.bfd);
 	close(esme->wqueue.bfd.fd);
 	esme->wqueue.bfd.fd = -1;
+	if (esme->acl)
+		esme->acl->esme = NULL;
 	smpp_esme_put(esme);
 
 	return 0;
@@ -885,6 +895,8 @@ static int esme_link_write_cb(struct osmo_fd *ofd, struct msgb *msg)
 		osmo_fd_unregister(&esme->wqueue.bfd);
 		close(esme->wqueue.bfd.fd);
 		esme->wqueue.bfd.fd = -1;
+		if (esme->acl)
+			esme->acl->esme = NULL;
 		smpp_esme_put(esme);
 	} else if (rc < msgb_length(msg)) {
 		LOGP(DSMPP, LOGL_ERROR, "[%s] Short write\n", esme->system_id);
@@ -965,14 +977,19 @@ struct smsc *smpp_smsc_alloc_init(void *ctx)
 /*! \brief Set the SMPP address and port without binding. */
 int smpp_smsc_conf(struct smsc *smsc, const char *bind_addr, uint16_t port)
 {
+	smsc->listen_port = port;
+
+	/* Avoid use-after-free if bind_addr == smsc->bind_addr */
+	if (smsc->bind_addr == bind_addr)
+		return 0;
+
 	talloc_free((void*)smsc->bind_addr);
 	smsc->bind_addr = NULL;
 	if (bind_addr) {
-		smsc->bind_addr = talloc_strdup(smsc, bind_addr);
+		smsc->bind_addr = bind_addr ? talloc_strdup(smsc, bind_addr) : NULL;
 		if (!smsc->bind_addr)
 			return -ENOMEM;
 	}
-	smsc->listen_port = port;
 	return 0;
 }
 
