@@ -20,9 +20,9 @@
  */
 
 #include <osmocom/core/fsm.h>
-#include <osmocom/gsm/gsup.h>
-#include <openbsc/vlr.h>
-#include <openbsc/debug.h>
+#include <osmocom/gsm/gsm48.h>
+#include <osmocom/msc/vlr.h>
+#include <osmocom/msc/debug.h>
 
 #include "vlr_core.h"
 #include "vlr_auth_fsm.h"
@@ -68,15 +68,20 @@ static const struct value_string upd_hlr_vlr_event_names[] = {
 	{ 0, NULL }
 };
 
+static inline struct vlr_subscr *upd_hlr_vlr_fi_priv(struct osmo_fsm_inst *fi);
+
 static void upd_hlr_vlr_fsm_init(struct osmo_fsm_inst *fi, uint32_t event,
 				 void *data)
 {
-	struct vlr_subscr *vsub = fi->priv;
+	struct vlr_subscr *vsub = upd_hlr_vlr_fi_priv(fi);
+	int rc;
 
 	OSMO_ASSERT(event == UPD_HLR_VLR_E_START);
 
 	/* Send UpdateLocation to HLR */
-	vlr_subscr_req_lu(vsub, vsub->vlr->cfg.is_ps);
+	rc = vlr_subscr_req_lu(vsub, vsub->vlr->cfg.is_ps);
+	if (rc < 0)
+		LOGPFSML(fi, LOGL_ERROR, "Failed to send UpdateLocation to HLR\n");
 	osmo_fsm_inst_state_chg(fi, UPD_HLR_VLR_S_WAIT_FOR_DATA,
 				LU_TIMEOUT_LONG, 0);
 }
@@ -84,7 +89,7 @@ static void upd_hlr_vlr_fsm_init(struct osmo_fsm_inst *fi, uint32_t event,
 static void upd_hlr_vlr_fsm_wait_data(struct osmo_fsm_inst *fi, uint32_t event,
 				      void *data)
 {
-	struct vlr_subscr *vsub = fi->priv;
+	struct vlr_subscr *vsub = upd_hlr_vlr_fi_priv(fi);
 
 	switch (event) {
 	case UPD_HLR_VLR_E_INS_SUB_DATA:
@@ -148,6 +153,12 @@ static struct osmo_fsm upd_hlr_vlr_fsm = {
 	.event_names = upd_hlr_vlr_event_names,
 };
 
+static inline struct vlr_subscr *upd_hlr_vlr_fi_priv(struct osmo_fsm_inst *fi)
+{
+	OSMO_ASSERT(fi->fsm == &upd_hlr_vlr_fsm);
+	return (struct vlr_subscr*)fi->priv;
+}
+
 struct osmo_fsm_inst *
 upd_hlr_vlr_proc_start(struct osmo_fsm_inst *parent,
 		       struct vlr_subscr *vsub,
@@ -190,10 +201,12 @@ static const struct value_string sub_pres_vlr_event_names[] = {
 	{ 0, NULL }
 };
 
+static inline struct vlr_subscr *sub_pres_vlr_fi_priv(struct osmo_fsm_inst *fi);
+
 static void sub_pres_vlr_fsm_init(struct osmo_fsm_inst *fi, uint32_t event,
 				  void *data)
 {
-	struct vlr_subscr *vsub = fi->priv;
+	struct vlr_subscr *vsub = sub_pres_vlr_fi_priv(fi);
 	OSMO_ASSERT(event == SUB_PRES_VLR_E_START);
 
 	if (!vsub->ms_not_reachable_flag) {
@@ -209,7 +222,7 @@ static void sub_pres_vlr_fsm_init(struct osmo_fsm_inst *fi, uint32_t event,
 static void sub_pres_vlr_fsm_wait_hlr(struct osmo_fsm_inst *fi, uint32_t event,
 				      void *data)
 {
-	struct vlr_subscr *vsub = fi->priv;
+	struct vlr_subscr *vsub = sub_pres_vlr_fi_priv(fi);
 
 	switch (event) {
 	case SUB_PRES_VLR_E_READY_SM_CNF:
@@ -252,21 +265,35 @@ static struct osmo_fsm sub_pres_vlr_fsm = {
 	.event_names = sub_pres_vlr_event_names,
 };
 
-struct osmo_fsm_inst *sub_pres_vlr_fsm_start(struct osmo_fsm_inst *parent,
-					     struct vlr_subscr *vsub,
-					     uint32_t term_event)
+static inline struct vlr_subscr *sub_pres_vlr_fi_priv(struct osmo_fsm_inst *fi)
+{
+	OSMO_ASSERT(fi->fsm == &sub_pres_vlr_fsm);
+	return (struct vlr_subscr*)fi->priv;
+}
+
+/* Note that the start event is dispatched right away, so in case the FSM immediately concludes from that
+ * event, the created FSM struct may no longer be valid as it already deallocated again, and it may
+ * furthermore already have invoked the parent FSM instance's deallocation as well. Hence, instead of
+ * returning, store the created FSM instance address in *fi_p before dispatching the event. It is thus
+ * possible to store the instance's pointer in a parent FSM instance without running danger of using
+ * already freed memory. */
+void sub_pres_vlr_fsm_start(struct osmo_fsm_inst **fi_p,
+			    struct osmo_fsm_inst *parent,
+			    struct vlr_subscr *vsub,
+			    uint32_t term_event)
 {
 	struct osmo_fsm_inst *fi;
 
+	OSMO_ASSERT(fi_p);
+
 	fi = osmo_fsm_inst_alloc_child(&sub_pres_vlr_fsm, parent,
 					term_event);
+	*fi_p = fi;
 	if (!fi)
-		return NULL;
+		return;
 
 	fi->priv = vsub;
 	osmo_fsm_inst_dispatch(fi, SUB_PRES_VLR_E_START, NULL);
-
-	return fi;
 }
 
 /***********************************************************************
@@ -311,11 +338,13 @@ struct lu_compl_vlr_priv {
 	bool assign_tmsi;
 };
 
+static inline struct lu_compl_vlr_priv *lu_compl_vlr_fi_priv(struct osmo_fsm_inst *fi);
+
 static void _vlr_lu_compl_fsm_done(struct osmo_fsm_inst *fi,
 				   enum vlr_fsm_result result,
 				   uint8_t cause)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	lcvp->result = result;
 	lcvp->cause = cause;
 	osmo_fsm_inst_state_chg(fi, LU_COMPL_VLR_S_DONE, 0, 0);
@@ -323,7 +352,7 @@ static void _vlr_lu_compl_fsm_done(struct osmo_fsm_inst *fi,
 
 static void vlr_lu_compl_fsm_success(struct osmo_fsm_inst *fi)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	struct vlr_subscr *vsub = lcvp->vsub;
 	if (!vsub->lu_complete) {
 		vsub->lu_complete = true;
@@ -335,7 +364,7 @@ static void vlr_lu_compl_fsm_success(struct osmo_fsm_inst *fi)
 
 static void vlr_lu_compl_fsm_failure(struct osmo_fsm_inst *fi, uint8_t cause)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	lcvp->vsub->vlr->ops.tx_lu_rej(lcvp->msc_conn_ref, cause);
 	_vlr_lu_compl_fsm_done(fi, VLR_FSM_RESULT_FAILURE, cause);
 }
@@ -343,7 +372,7 @@ static void vlr_lu_compl_fsm_failure(struct osmo_fsm_inst *fi, uint8_t cause)
 static void vlr_lu_compl_fsm_dispatch_result(struct osmo_fsm_inst *fi,
 					     uint32_t prev_state)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	if (!fi->proc.parent) {
 		LOGPFSML(fi, LOGL_ERROR, "No parent FSM\n");
 		return;
@@ -358,7 +387,7 @@ static void vlr_lu_compl_fsm_dispatch_result(struct osmo_fsm_inst *fi,
 static void lu_compl_vlr_init(struct osmo_fsm_inst *fi, uint32_t event,
 			      void *data)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	struct vlr_subscr *vsub = lcvp->vsub;
 	struct vlr_instance *vlr;
 	OSMO_ASSERT(vsub);
@@ -384,14 +413,12 @@ static void lu_compl_vlr_init(struct osmo_fsm_inst *fi, uint32_t event,
 	osmo_fsm_inst_state_chg(fi, LU_COMPL_VLR_S_WAIT_SUB_PRES,
 				LU_TIMEOUT_LONG, 0);
 
-	lcvp->sub_pres_vlr_fsm = sub_pres_vlr_fsm_start(fi, vsub,
-						LU_COMPL_VLR_E_SUB_PRES_COMPL);
-
+	sub_pres_vlr_fsm_start(&lcvp->sub_pres_vlr_fsm, fi, vsub, LU_COMPL_VLR_E_SUB_PRES_COMPL);
 }
 
 static void lu_compl_vlr_new_tmsi(struct osmo_fsm_inst *fi)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	struct vlr_subscr *vsub = lcvp->vsub;
 	struct vlr_instance *vlr = vsub->vlr;
 
@@ -414,7 +441,7 @@ static void lu_compl_vlr_wait_subscr_pres(struct osmo_fsm_inst *fi,
 					  uint32_t event,
 					  void *data)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	struct vlr_subscr *vsub = lcvp->vsub;
 	struct vlr_instance *vlr = vsub->vlr;
 
@@ -450,7 +477,7 @@ static void lu_compl_vlr_wait_subscr_pres(struct osmo_fsm_inst *fi,
 static void lu_compl_vlr_wait_imei(struct osmo_fsm_inst *fi, uint32_t event,
 				   void *data)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	struct vlr_subscr *vsub = lcvp->vsub;
 	struct vlr_instance *vlr = vsub->vlr;
 
@@ -492,11 +519,13 @@ static void lu_compl_vlr_wait_imei(struct osmo_fsm_inst *fi, uint32_t event,
 	vlr_lu_compl_fsm_success(fi);
 }
 
+static inline struct lu_compl_vlr_priv *lu_compl_vlr_fi_priv(struct osmo_fsm_inst *fi);
+
 /* Waiting for TMSI confirmation */
 static void lu_compl_vlr_wait_tmsi(struct osmo_fsm_inst *fi, uint32_t event,
 				   void *data)
 {
-	struct lu_compl_vlr_priv *lcvp = fi->priv;
+	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
 	struct vlr_subscr *vsub = lcvp->vsub;
 
 	OSMO_ASSERT(event == LU_COMPL_VLR_E_NEW_TMSI_ACK);
@@ -570,6 +599,12 @@ static struct osmo_fsm lu_compl_vlr_fsm = {
 	.event_names = lu_compl_vlr_event_names,
 };
 
+static inline struct lu_compl_vlr_priv *lu_compl_vlr_fi_priv(struct osmo_fsm_inst *fi)
+{
+	OSMO_ASSERT(fi->fsm == &lu_compl_vlr_fsm);
+	return (struct lu_compl_vlr_priv*)fi->priv;
+}
+
 struct osmo_fsm_inst *
 lu_compl_vlr_proc_alloc(struct osmo_fsm_inst *parent,
 			struct vlr_subscr *vsub,
@@ -638,7 +673,7 @@ struct lu_fsm_priv {
 	struct osmo_location_area_id old_lai;
 	struct osmo_location_area_id new_lai;
 	bool authentication_required;
-	enum vlr_ciph ciphering_required;
+	bool ciphering_required;
 	bool is_r99;
 	bool is_utran;
 	bool assign_tmsi;
@@ -659,14 +694,13 @@ static bool is_auth_required(struct lu_fsm_priv *lfp)
 	/* The cases where the authentication procedure should be used
 	 * are defined in 3GPP TS 33.102 */
 	/* For now we use a default value passed in to vlr_lu_fsm(). */
-	return lfp->authentication_required
-	       || (lfp->ciphering_required != VLR_CIPH_NONE);
+	return lfp->authentication_required || lfp->ciphering_required;
 }
 
 /* Determine if ciphering is required */
 static bool is_ciph_required(struct lu_fsm_priv *lfp)
 {
-	return lfp->ciphering_required != VLR_CIPH_NONE;
+	return lfp->ciphering_required;
 }
 
 /* Determine if a HLR Update is required */
@@ -677,10 +711,12 @@ static bool hlr_update_needed(struct vlr_subscr *vsub)
 	return true;
 }
 
+static inline struct lu_fsm_priv *lu_fsm_fi_priv(struct osmo_fsm_inst *fi);
+
 static void lu_fsm_dispatch_result(struct osmo_fsm_inst *fi,
 				   uint32_t prev_state)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	if (!fi->proc.parent) {
 		LOGPFSML(fi, LOGL_ERROR, "No parent FSM\n");
 		return;
@@ -695,7 +731,7 @@ static void lu_fsm_dispatch_result(struct osmo_fsm_inst *fi,
 static void _lu_fsm_done(struct osmo_fsm_inst *fi,
 			 enum vlr_fsm_result result)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	lfp->result = result;
 	osmo_fsm_inst_state_chg(fi, VLR_ULA_S_DONE, 0, 0);
 }
@@ -705,17 +741,16 @@ static void lu_fsm_success(struct osmo_fsm_inst *fi)
 	_lu_fsm_done(fi, VLR_FSM_RESULT_SUCCESS);
 }
 
-static void lu_fsm_failure(struct osmo_fsm_inst *fi, uint8_t rej_cause)
+static void lu_fsm_failure(struct osmo_fsm_inst *fi, enum gsm48_reject_value rej_cause)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
-	if (rej_cause)
-		lfp->vlr->ops.tx_lu_rej(lfp->msc_conn_ref, rej_cause);
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
+	lfp->vlr->ops.tx_lu_rej(lfp->msc_conn_ref, rej_cause ? : GSM48_REJECT_NETWORK_FAILURE);
 	_lu_fsm_done(fi, VLR_FSM_RESULT_FAILURE);
 }
 
 static void vlr_loc_upd_start_lu_compl_fsm(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	lfp->lu_compl_vlr_fsm =
 		lu_compl_vlr_proc_alloc(fi, lfp->vsub, lfp->msc_conn_ref,
 					VLR_ULA_E_LU_COMPL_SUCCESS,
@@ -727,7 +762,7 @@ static void vlr_loc_upd_start_lu_compl_fsm(struct osmo_fsm_inst *fi)
 
 static void lu_fsm_discard_lu_compl_fsm(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	if (!lfp->lu_compl_vlr_fsm)
 		return;
 	osmo_fsm_inst_term(lfp->lu_compl_vlr_fsm, OSMO_FSM_TERM_PARENT, NULL);
@@ -736,7 +771,7 @@ static void lu_fsm_discard_lu_compl_fsm(struct osmo_fsm_inst *fi)
 /* 4.1.2.1 Node 4 */
 static void vlr_loc_upd_node_4(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_subscr *vsub = lfp->vsub;
 	bool hlr_unknown = false;
 
@@ -776,7 +811,7 @@ static void vlr_loc_upd_node_b(struct osmo_fsm_inst *fi)
 /* Non-standard: after Ciphering Mode Complete (or no ciph required) */
 static void vlr_loc_upd_post_ciph(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_subscr *vsub = lfp->vsub;
 
 	LOGPFSM(fi, "%s()\n", __func__);
@@ -808,8 +843,9 @@ static void vlr_loc_upd_post_ciph(struct osmo_fsm_inst *fi)
 /* 4.1.2.1 after Authentication successful (or no auth rqd) */
 static void vlr_loc_upd_post_auth(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_subscr *vsub = lfp->vsub;
+	bool umts_aka;
 
 	LOGPFSM(fi, "%s()\n", __func__);
 
@@ -820,12 +856,32 @@ static void vlr_loc_upd_post_auth(struct osmo_fsm_inst *fi)
 		return;
 	}
 
+	if (!vsub->last_tuple) {
+		LOGPFSML(fi, LOGL_ERROR, "No auth tuple available\n");
+		lu_fsm_failure(fi, GSM48_REJECT_NETWORK_FAILURE);
+		return;
+	}
+
+	switch (vsub->sec_ctx) {
+	case VLR_SEC_CTX_GSM:
+		umts_aka = false;
+		break;
+	case VLR_SEC_CTX_UMTS:
+		umts_aka = true;
+		break;
+	default:
+		LOGPFSML(fi, LOGL_ERROR, "Cannot start ciphering, security context is not established\n");
+		lu_fsm_failure(fi, GSM48_REJECT_NETWORK_FAILURE);
+		return;
+	}
+
 	if (vlr_set_ciph_mode(vsub->vlr, fi, lfp->msc_conn_ref,
 			      lfp->ciphering_required,
+			      umts_aka,
 			      vsub->vlr->cfg.retrieve_imeisv_ciphered)) {
 		LOGPFSML(fi, LOGL_ERROR,
 			 "Failed to send Ciphering Mode Command\n");
-		vlr_lu_compl_fsm_failure(fi, GSM48_REJECT_NETWORK_FAILURE);
+		lu_fsm_failure(fi, GSM48_REJECT_NETWORK_FAILURE);
 		return;
 	}
 
@@ -834,7 +890,7 @@ static void vlr_loc_upd_post_auth(struct osmo_fsm_inst *fi)
 
 static void vlr_loc_upd_node1(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_subscr *vsub = lfp->vsub;
 
 	LOGPFSM(fi, "%s()\n", __func__);
@@ -857,7 +913,7 @@ static void vlr_loc_upd_node1(struct osmo_fsm_inst *fi)
 
 static void vlr_loc_upd_want_imsi(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_instance *vlr = lfp->vlr;
 
 	LOGPFSM(fi, "%s()\n", __func__);
@@ -873,7 +929,7 @@ static void vlr_loc_upd_want_imsi(struct osmo_fsm_inst *fi)
 
 static int assoc_lfp_with_sub(struct osmo_fsm_inst *fi, struct vlr_subscr *vsub)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_instance *vlr = lfp->vlr;
 
 	if (vsub->lu_fsm) {
@@ -896,17 +952,9 @@ static int assoc_lfp_with_sub(struct osmo_fsm_inst *fi, struct vlr_subscr *vsub)
 	return 0;
 }
 
-static const char *lai_name(struct osmo_location_area_id *lai)
-{
-	static char buf[64];
-	snprintf(buf, sizeof(buf),"MCC:%u, MNC:%u, LAC:%u",
-		 lai->plmn.mcc, lai->plmn.mnc, lai->lac);
-	return buf;
-}
-
 static int _lu_fsm_associate_vsub(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_instance *vlr = lfp->vlr;
 	struct vlr_subscr *vsub = NULL;
 
@@ -958,7 +1006,7 @@ static int _lu_fsm_associate_vsub(struct osmo_fsm_inst *fi)
 /* 4.1.2.1: Subscriber (via MSC/SGSN) requests location update */
 static void _start_lu_main(struct osmo_fsm_inst *fi)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_instance *vlr = lfp->vlr;
 
 	/* TODO: PUESBINE related handling */
@@ -973,7 +1021,7 @@ static void _start_lu_main(struct osmo_fsm_inst *fi)
 #endif
 		LOGPFSML(fi, LOGL_NOTICE, "LAI change from %s,"
 			 " but checking previous VLR not implemented\n",
-			 lai_name(&lfp->old_lai));
+			 osmo_lai_name(&lfp->old_lai));
 	}
 
 	/* If this is a TMSI based LU, we may not have the IMSI. Make sure that
@@ -987,7 +1035,7 @@ static void _start_lu_main(struct osmo_fsm_inst *fi)
 static void lu_fsm_idle(struct osmo_fsm_inst *fi, uint32_t event,
 			void *data)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_instance *vlr = lfp->vlr;
 
 	OSMO_ASSERT(event == VLR_ULA_E_UPDATE_LA);
@@ -1047,51 +1095,26 @@ static void lu_fsm_wait_pvlr(struct osmo_fsm_inst *fi, uint32_t event,
 static void lu_fsm_wait_auth(struct osmo_fsm_inst *fi, uint32_t event,
 			     void *data)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
-	enum vlr_auth_fsm_result *res = data;
-	uint8_t rej_cause = 0;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
+	enum gsm48_reject_value *res = data;
 
 	OSMO_ASSERT(event == VLR_ULA_E_AUTH_RES);
 
 	lfp->upd_hlr_vlr_fsm = NULL;
 
-	if (res) {
-		switch (*res) {
-		case VLR_AUTH_RES_PASSED:
-			/* Result == Pass */
-			vlr_loc_upd_post_auth(fi);
-			return;
-		case VLR_AUTH_RES_ABORTED:
-			/* go to Idle with no response */
-			rej_cause = 0;
-			break;
-		case VLR_AUTH_RES_UNKNOWN_SUBSCR:
-			/* FIXME: delete subscribe record */
-			rej_cause = GSM48_REJECT_IMSI_UNKNOWN_IN_HLR;
-			break;
-		case VLR_AUTH_RES_AUTH_FAILED:
-			/* cause = illegal subscriber */
-			rej_cause = GSM48_REJECT_ILLEGAL_MS;
-			break;
-		case VLR_AUTH_RES_PROC_ERR:
-			/* cause = system failure */
-			rej_cause = GSM48_REJECT_NETWORK_FAILURE;
-			break;
-		default:
-			LOGPFSML(fi, LOGL_ERROR, "event without effect: %s\n",
-				 osmo_fsm_event_name(fi->fsm, event));
-			break;
-		}
-	} else
-		rej_cause = GSM48_REJECT_NETWORK_FAILURE;
+	if (!res || *res) {
+		lu_fsm_failure(fi, res? *res : GSM48_REJECT_NETWORK_FAILURE);
+		return;
+	}
 
-	lu_fsm_failure(fi, rej_cause);
+	/* Result == Pass */
+	vlr_loc_upd_post_auth(fi);
 }
 
 static void lu_fsm_wait_ciph(struct osmo_fsm_inst *fi, uint32_t event,
 			     void *data)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_subscr *vsub = lfp->vsub;
 	struct vlr_ciph_result res = { .cause = VLR_CIPH_REJECT };
 
@@ -1116,7 +1139,7 @@ static void lu_fsm_wait_ciph(struct osmo_fsm_inst *fi, uint32_t event,
 		return;
 	}
 
-	if (res.imeisv) {
+	if (*res.imeisv) {
 		LOGPFSM(fi, "got IMEISV: %s\n", res.imeisv);
 		vlr_subscr_set_imeisv(vsub, res.imeisv);
 	}
@@ -1126,7 +1149,7 @@ static void lu_fsm_wait_ciph(struct osmo_fsm_inst *fi, uint32_t event,
 static void lu_fsm_wait_imsi(struct osmo_fsm_inst *fi, uint32_t event,
 			     void *data)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_subscr *vsub = lfp->vsub;
 	char *mi_string = data;
 
@@ -1146,7 +1169,7 @@ static void lu_fsm_wait_imsi(struct osmo_fsm_inst *fi, uint32_t event,
 static void lu_fsm_wait_hlr_ul_res(struct osmo_fsm_inst *fi, uint32_t event,
 				   void *data)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 
 	switch (event) {
 	case VLR_ULA_E_HLR_LU_RES:
@@ -1189,7 +1212,7 @@ static void lu_fsm_wait_hlr_ul_res(struct osmo_fsm_inst *fi, uint32_t event,
 static void lu_fsm_wait_lu_compl(struct osmo_fsm_inst *fi, uint32_t event,
 				 void *data)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	uint8_t cause;
 
 	switch (event) {
@@ -1208,6 +1231,10 @@ static void lu_fsm_wait_lu_compl(struct osmo_fsm_inst *fi, uint32_t event,
 		/* TODO: Set_Notification_Type 23.078 */
 		/* TODO: Notify_gsmSCF 23.078 */
 		/* TODO: Authenticated Radio Contact Established -> ARC */
+
+		if (lfp->type == VLR_LU_TYPE_IMSI_ATTACH)
+			lfp->vlr->ops.tx_mm_info(lfp->msc_conn_ref);
+
 		lu_fsm_success(fi);
 		break;
 	case VLR_ULA_E_LU_COMPL_FAILURE:
@@ -1228,7 +1255,7 @@ static void lu_fsm_wait_lu_compl(struct osmo_fsm_inst *fi, uint32_t event,
 static void lu_fsm_wait_lu_compl_standalone(struct osmo_fsm_inst *fi,
 					uint32_t event, void *data)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_subscr *vsub = lfp->vsub;
 	uint8_t cause;
 
@@ -1348,7 +1375,7 @@ static const struct osmo_fsm_state vlr_lu_fsm_states[] = {
 
 static void fsm_lu_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
-	struct lu_fsm_priv *lfp = fi->priv;
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
 	struct vlr_subscr *vsub = lfp->vsub;
 
 	LOGPFSM(fi, "fsm_lu_cleanup called with cause %s\n",
@@ -1368,6 +1395,12 @@ static struct osmo_fsm vlr_lu_fsm = {
 	.cleanup = fsm_lu_cleanup,
 };
 
+static inline struct lu_fsm_priv *lu_fsm_fi_priv(struct osmo_fsm_inst *fi)
+{
+	OSMO_ASSERT(fi->fsm == &vlr_lu_fsm);
+	return (struct lu_fsm_priv*)fi->priv;
+}
+
 struct osmo_fsm_inst *
 vlr_loc_update(struct osmo_fsm_inst *parent,
 	       uint32_t parent_event_success,
@@ -1378,7 +1411,7 @@ vlr_loc_update(struct osmo_fsm_inst *parent,
 	       const struct osmo_location_area_id *old_lai,
 	       const struct osmo_location_area_id *new_lai,
 	       bool authentication_required,
-	       enum vlr_ciph ciphering_required,
+	       bool ciphering_required,
 	       bool is_r99, bool is_utran,
 	       bool assign_tmsi)
 {
@@ -1430,14 +1463,20 @@ vlr_loc_update(struct osmo_fsm_inst *parent,
 	return fi;
 }
 
-/* Gracefully terminate an FSM created by vlr_loc_update() in case of external
- * timeout (i.e. from MSC). */
-void vlr_loc_update_conn_timeout(struct osmo_fsm_inst *fi)
+void vlr_loc_update_cancel(struct osmo_fsm_inst *fi,
+			   enum osmo_fsm_term_cause fsm_cause,
+			   uint8_t gsm48_cause)
 {
-	if (!fi || fi->state == VLR_ULA_S_DONE)
-		return;
-	LOGPFSM(fi, "Connection timed out\n");
-	lu_fsm_failure(fi, GSM48_REJECT_CONGESTION);
+	struct lu_fsm_priv *lfp;
+
+	OSMO_ASSERT(fi);
+	OSMO_ASSERT(fi->fsm == &vlr_lu_fsm);
+
+	lfp = fi->priv;
+	lfp->rej_cause = gsm48_cause;
+
+	if (fi->state != VLR_ULA_S_DONE)
+		lu_fsm_failure(fi, gsm48_cause);
 }
 
 void vlr_lu_fsm_init(void)
