@@ -30,20 +30,27 @@
 #include <osmocom/core/application.h>
 #include <osmocom/gsm/protocol/gsm_04_11.h>
 #include <osmocom/gsm/gsup.h>
+#include <osmocom/msc/gsup_client.h>
+#include <osmocom/msc/gsm_04_11.h>
+#include <osmocom/msc/debug.h>
+#include <osmocom/msc/gsm_04_08.h>
+#include <osmocom/msc/transaction.h>
+
+#if BUILD_IU
+#include <osmocom/msc/iucs_ranap.h>
 #include <osmocom/ranap/iu_client.h>
-#include <openbsc/gsup_client.h>
-#include <openbsc/gsm_04_11.h>
-#include <openbsc/debug.h>
-#include <openbsc/iucs_ranap.h>
+#else
+#include <osmocom/msc/iu_dummy.h>
+#endif
 
 #include "msc_vlr_tests.h"
 
+void *msc_vlr_tests_ctx = NULL;
+
 bool _log_lines = false;
 
 struct gsm_network *net = NULL;
 
-struct gsm_bts *the_bts;
-
 const char *gsup_tx_expected = NULL;
 bool gsup_tx_confirmed;
 
@@ -57,12 +64,51 @@ const char *auth_request_expect_rand;
 const char *auth_request_expect_autn;
 bool cipher_mode_cmd_sent;
 bool cipher_mode_cmd_sent_with_imeisv;
+const char *cipher_mode_expect_kc;
+bool security_mode_ctrl_sent;
+const char *security_mode_expect_ck;
+const char *security_mode_expect_ik;
 
 bool iu_release_expected = false;
 bool iu_release_sent = false;
 bool bssap_clear_expected = false;
 bool bssap_clear_sent = false;
 
+uint32_t cc_to_mncc_tx_expected_msg_type = 0;
+const char *cc_to_mncc_tx_expected_imsi = NULL;
+bool cc_to_mncc_tx_confirmed = false;
+uint32_t cc_to_mncc_tx_got_callref = 0;
+
+extern int gsm0407_pdisc_ctr_bin(uint8_t pdisc);
+
+/* static state variables for the L3 send sequence numbers */
+static uint8_t n_sd[4];
+
+/* patch a correct send sequence number into the given message */
+static void patch_l3_seq_nr(struct msgb *msg)
+{
+	struct gsm48_hdr *gh = msgb_l3(msg);
+	uint8_t pdisc = gsm48_hdr_pdisc(gh);
+	uint8_t *msg_type_oct = &msg->l3h[1];
+	int bin = gsm0407_pdisc_ctr_bin(pdisc);
+
+	if (bin >= 0 && bin < ARRAY_SIZE(n_sd)) {
+		/* patch in n_sd into the msg_type octet */
+		*msg_type_oct = (*msg_type_oct & 0x3f) | ((n_sd[bin] & 0x3) << 6);
+		//fprintf(stderr, "pdisc=0x%02x bin=%d, patched n_sd=%u\n\n", pdisc, bin, n_sd[bin] & 3);
+		/* increment N(SD) */
+		n_sd[bin] = (n_sd[bin] + 1) % 4;
+	} else {
+		//fprintf(stderr, "pdisc=0x%02x NO SEQ\n\n", pdisc);
+	}
+}
+
+/* reset L3 sequence numbers (e.g. new RR connection) */
+static void reset_l3_seq_nr()
+{
+	memset(n_sd, 0, sizeof(n_sd));
+}
+
 struct msgb *msgb_from_hex(const char *label, uint16_t size, const char *hex)
 {
 	struct msgb *msg = msgb_alloc(size, label);
@@ -73,17 +119,12 @@ struct msgb *msgb_from_hex(const char *label, uint16_t size, const char *hex)
 	return msg;
 }
 
-const char *gh_type_name(struct gsm48_hdr *gh)
+static const char *gh_type_name(struct gsm48_hdr *gh)
 {
 	return gsm48_pdisc_msgtype_name(gsm48_hdr_pdisc(gh),
 					gsm48_hdr_msg_type(gh));
 }
 
-const char *msg_type_name(struct msgb *msg)
-{
-	return gh_type_name((void*)msg->data);
-}
-
 void dtap_expect_tx(const char *hex)
 {
 	/* Has the previously expected dtap been received? */
@@ -91,6 +132,9 @@ void dtap_expect_tx(const char *hex)
 	if (!hex)
 		return;
 	dtap_tx_expected = msgb_from_hex("dtap_tx_expected", 1024, hex);
+	/* Mask the sequence number out */
+	if (msgb_length(dtap_tx_expected) >= 2)
+		dtap_tx_expected->data[1] &= 0x3f;
 	dtap_tx_confirmed = false;
 }
 
@@ -124,12 +168,12 @@ void gsup_rx(const char *rx_hex, const char *expect_tx_hex)
 	label = osmo_gsup_message_type_name(msg->l2h[0]);
 	fprintf(stderr, "<-- GSUP rx %s: %s\n", label,
 		osmo_hexdump_nospc(msgb_l2(msg), msgb_l2len(msg)));
+	/* GSUP read cb takes ownership of msgb */
 	rc = vlr_gsupc_read_cb(net->vlr->gsup_client, msg);
 	fprintf(stderr, "<-- GSUP rx %s: vlr_gsupc_read_cb() returns %d\n",
 		label, rc);
 	if (expect_tx_hex)
 		OSMO_ASSERT(gsup_tx_confirmed);
-	talloc_free(msg);
 }
 
 bool conn_exists(struct gsm_subscriber_connection *conn)
@@ -147,9 +191,7 @@ enum ran_type rx_from_ran = RAN_GERAN_A;
 struct gsm_subscriber_connection *conn_new(void)
 {
 	struct gsm_subscriber_connection *conn;
-	conn = msc_subscr_con_allocate(net);
-	conn->bts = the_bts;
-	conn->via_ran = rx_from_ran;
+	conn = msc_subscr_conn_alloc(net, rx_from_ran, 23);
 	if (conn->via_ran == RAN_UTRAN_IU) {
 		struct ranap_ue_conn_ctx *ue_ctx = talloc_zero(conn, struct ranap_ue_conn_ctx);
 		*ue_ctx = (struct ranap_ue_conn_ctx){
@@ -164,8 +206,6 @@ struct gsm_subscriber_connection *g_conn = NULL;
 
 void rx_from_ms(struct msgb *msg)
 {
-	int rc;
-
 	struct gsm48_hdr *gh = msgb_l3(msg);
 
 	log("MSC <--%s-- MS: %s",
@@ -178,12 +218,11 @@ void rx_from_ms(struct msgb *msg)
 	if (!g_conn) {
 		log("new conn");
 		g_conn = conn_new();
-		rc = msc_compl_l3(g_conn, msg, 23);
-		if (rc == BSC_API_CONN_POL_REJECT) {
-			msc_subscr_con_free(g_conn);
-			g_conn = NULL;
-		}
+		reset_l3_seq_nr();
+		patch_l3_seq_nr(msg);
+		msc_compl_l3(g_conn, msg, 23);
 	} else {
+		patch_l3_seq_nr(msg);
 		if ((gsm48_hdr_pdisc(gh) == GSM48_PDISC_RR)
 		    && (gsm48_hdr_msg_type(gh) == GSM48_MT_RR_CIPH_M_COMPL))
 			msc_cipher_mode_compl(g_conn, msg, 0);
@@ -205,7 +244,7 @@ void ms_sends_msg(const char *hex)
 	talloc_free(msg);
 }
 
-int ms_sends_msg_fake(uint8_t pdisc, uint8_t msg_type)
+static int ms_sends_msg_fake(uint8_t pdisc, uint8_t msg_type)
 {
 	int rc;
 	struct msgb *msg;
@@ -220,35 +259,45 @@ int ms_sends_msg_fake(uint8_t pdisc, uint8_t msg_type)
 	/* some amount of data, whatever */
 	msgb_put(msg, 123);
 
+	patch_l3_seq_nr(msg);
 	rc = gsm0408_dispatch(g_conn, msg);
 
 	talloc_free(msg);
 	return rc;
 }
 
+static inline void ms_msg_log_err(uint8_t val, uint8_t msgtype)
+{
+	int rc = ms_sends_msg_fake(val, msgtype);
+	if (rc != -EACCES)
+		log("Unexpected return value %u != %u for %s/%s",
+		    -rc, -EACCES, gsm48_pdisc_name(val), gsm48_cc_msg_name(msgtype));
+}
+
 void thwart_rx_non_initial_requests()
 {
 	log("requests shall be thwarted");
-	OSMO_ASSERT(ms_sends_msg_fake(GSM48_PDISC_CC, GSM48_MT_CC_SETUP) == -EACCES);
-	OSMO_ASSERT(ms_sends_msg_fake(GSM48_PDISC_MM, 0x33 /* nonexistent */) == -EACCES);
-	OSMO_ASSERT(ms_sends_msg_fake(GSM48_PDISC_RR, GSM48_MT_RR_SYSINFO_1) == -EACCES);
-	OSMO_ASSERT(ms_sends_msg_fake(GSM48_PDISC_SMS, GSM411_MT_CP_DATA) == -EACCES);
+
+	ms_msg_log_err(GSM48_PDISC_CC, GSM48_MT_CC_SETUP);
+	ms_msg_log_err(GSM48_PDISC_MM, 0x33); /* nonexistent */
+	ms_msg_log_err(GSM48_PDISC_RR, GSM48_MT_RR_SYSINFO_1);
+	ms_msg_log_err(GSM48_PDISC_SMS, GSM411_MT_CP_DATA);
 }
 
 void send_sms(struct vlr_subscr *receiver,
 	      struct vlr_subscr *sender,
 	      char *str)
 {
-	struct gsm_sms *sms = sms_from_text(receiver, sender, 0, str);
+	struct gsm_sms *sms = sms_from_text(receiver, sender->msisdn, 0, str);
 	gsm411_send_sms_subscr(receiver, sms);
 }
 
 unsigned char next_rand_byte = 0;
-/* override, requires '-Wl,--wrap=RAND_bytes' */
-int __real_RAND_bytes(unsigned char *buf, int num);
-int __wrap_RAND_bytes(unsigned char *buf, int num)
+/* override, requires '-Wl,--wrap=osmo_get_rand_id' */
+int __real_osmo_get_rand_id(uint8_t *buf, size_t num);
+int __wrap_osmo_get_rand_id(uint8_t *buf, size_t num)
 {
-	int i;
+	size_t i;
 	for (i = 0; i < num; i++)
 		buf[i] = next_rand_byte++;
 	return 1;
@@ -279,7 +328,7 @@ void paging_expect_tmsi(uint32_t tmsi)
 	paging_expecting_imsi = NULL;
 }
 
-int _paging_sent(enum ran_type via_ran, const char *imsi, uint32_t tmsi, uint32_t lac)
+static int _paging_sent(enum ran_type via_ran, const char *imsi, uint32_t tmsi, uint32_t lac)
 {
 	log("%s sends out paging request to IMSI %s, TMSI 0x%08x, LAC %u",
 	    ran_type_name(via_ran), imsi, tmsi, lac);
@@ -323,17 +372,27 @@ void clear_vlr()
 	}
 
 	net->authentication_required = false;
-	net->a5_encryption = VLR_CIPH_NONE;
+	net->a5_encryption_mask = (1 << 0);
 	net->vlr->cfg.check_imei_rqd = false;
 	net->vlr->cfg.assign_tmsi = false;
 	net->vlr->cfg.retrieve_imeisv_early = false;
 	net->vlr->cfg.retrieve_imeisv_ciphered = false;
+	net->vlr->cfg.auth_tuple_max_reuse_count = 0;
+	net->vlr->cfg.auth_reuse_old_sets_on_error = false;
 
 	rx_from_ran = RAN_GERAN_A;
 	auth_request_sent = false;
 	auth_request_expect_rand = NULL;
 	auth_request_expect_autn = NULL;
 
+	cipher_mode_cmd_sent = false;
+	cipher_mode_cmd_sent_with_imeisv = false;
+	cipher_mode_expect_kc = NULL;
+
+	security_mode_ctrl_sent = false;
+	security_mode_expect_ck = NULL;
+	security_mode_expect_ik = NULL;
+
 	next_rand_byte = 0;
 
 	iu_release_expected = false;
@@ -368,7 +427,7 @@ static struct log_info_cat test_categories[] = {
 	[DCC] = {
 		.name = "DCC",
 		.description = "Layer3 Call Control (CC)",
-		.enabled = 1, .loglevel = LOGL_NOTICE,
+		.enabled = 1, .loglevel = LOGL_DEBUG,
 	},
 	[DMM] = {
 		.name = "DMM",
@@ -395,6 +454,11 @@ static struct log_info_cat test_categories[] = {
 		.description = "Iu-CS Protocol",
 		.enabled = 1, .loglevel = LOGL_DEBUG,
 	},
+	[DMNCC] = {
+		.name = "DMNCC",
+		.description = "MNCC API for Call Control application",
+		.enabled = 1, .loglevel = LOGL_DEBUG,
+	},
 };
 
 static struct log_info info = {
@@ -402,11 +466,31 @@ static struct log_info info = {
 	.num_cat = ARRAY_SIZE(test_categories),
 };
 
-extern void *tall_bsc_ctx;
-
-int fake_mncc_recv(struct gsm_network *net, struct msgb *msg)
+int mncc_recv(struct gsm_network *net, struct msgb *msg)
 {
-	fprintf(stderr, "rx MNCC\n");
+	struct gsm_mncc *mncc = (void*)msg->data;
+	log("MSC --> MNCC: callref 0x%x: %s", mncc->callref,
+	    get_mncc_name(mncc->msg_type));
+
+	OSMO_ASSERT(cc_to_mncc_tx_expected_msg_type);
+	if (cc_to_mncc_tx_expected_msg_type != mncc->msg_type) {
+		log("Mismatch! Expected MNCC msg type: %s",
+		    get_mncc_name(cc_to_mncc_tx_expected_msg_type));
+		abort();
+	}
+
+	if (strcmp(cc_to_mncc_tx_expected_imsi, mncc->imsi)) {
+		log("Mismatch! Expected MNCC msg IMSI: '%s', got '%s'",
+		    cc_to_mncc_tx_expected_imsi,
+		    mncc->imsi);
+		abort();
+	}
+
+	cc_to_mncc_tx_confirmed = true;
+	cc_to_mncc_tx_got_callref = mncc->callref;
+	cc_to_mncc_tx_expected_imsi = NULL;
+	cc_to_mncc_tx_expected_msg_type = 0;
+	talloc_free(msg);
 	return 0;
 }
 
@@ -421,7 +505,7 @@ __wrap_gsup_client_create(const char *ip_addr, unsigned int tcp_port,
 			  struct oap_client_config *oap_config)
 {
 	struct gsup_client *gsupc;
-	gsupc = talloc_zero(tall_bsc_ctx, struct gsup_client);
+	gsupc = talloc_zero(msc_vlr_tests_ctx, struct gsup_client);
 	OSMO_ASSERT(gsupc);
 	return gsupc;
 }
@@ -446,13 +530,16 @@ int __wrap_gsup_client_send(struct gsup_client *gsupc, struct msgb *msg)
 	return 0;
 }
 
-int _validate_dtap(struct msgb *msg, enum ran_type to_ran)
+static int _validate_dtap(struct msgb *msg, enum ran_type to_ran)
 {
 	btw("DTAP --%s--> MS: %s: %s",
-	    ran_type_name(to_ran), msg_type_name(msg),
+	    ran_type_name(to_ran), gh_type_name((void*)msg->data),
 	    osmo_hexdump_nospc(msg->data, msg->len));
 
 	OSMO_ASSERT(dtap_tx_expected);
+
+	/* Mask the sequence number out before comparing */
+	msg->data[1] &= 0x3f;
 	if (msg->len != dtap_tx_expected->len
 	    || memcmp(msg->data, dtap_tx_expected->data, msg->len)) {
 		fprintf(stderr, "Mismatch! Expected:\n%s\n",
@@ -514,6 +601,30 @@ int __wrap_a_iface_tx_clear_cmd(struct gsm_subscriber_connection *conn)
 	return 0;
 }
 
+/* override, requires '-Wl,--wrap=msc_mgcp_call_assignment' */
+int __real_msc_mgcp_call_assignment(struct gsm_trans *trans);
+int __wrap_msc_mgcp_call_assignment(struct gsm_trans *trans)
+{
+	log("MS <--Call Assignment-- MSC: subscr=%s callref=0x%x",
+	    vlr_subscr_name(trans->vsub), trans->callref);
+	return 0;
+}
+
+struct gsm_mncc *on_call_release_mncc_sends_to_cc_data = NULL;
+
+/* override, requires '-Wl,--wrap=msc_mgcp_call_release' */
+void __real_msc_mgcp_call_release(struct gsm_trans *trans);
+void __wrap_msc_mgcp_call_release(struct gsm_trans *trans)
+{
+	log("MS <--Call Release-- MSC: subscr=%s callref=0x%x",
+	    vlr_subscr_name(trans->vsub), trans->callref);
+	if (on_call_release_mncc_sends_to_cc_data) {
+		mncc_tx_to_cc(trans->net, on_call_release_mncc_sends_to_cc_data->msg_type,
+			      on_call_release_mncc_sends_to_cc_data);
+		on_call_release_mncc_sends_to_cc_data = NULL;
+	}
+}
+
 static int fake_vlr_tx_lu_acc(void *msc_conn_ref, uint32_t send_tmsi)
 {
 	struct gsm_subscriber_connection *conn = msc_conn_ref;
@@ -526,7 +637,7 @@ static int fake_vlr_tx_lu_acc(void *msc_conn_ref, uint32_t send_tmsi)
 	return 0;
 }
 
-static int fake_vlr_tx_lu_rej(void *msc_conn_ref, uint8_t cause)
+static int fake_vlr_tx_lu_rej(void *msc_conn_ref, enum gsm48_reject_value cause)
 {
 	struct gsm_subscriber_connection *conn = msc_conn_ref;
 	btw("sending LU Reject for %s, cause %u", vlr_subscr_name(conn->vsub), cause);
@@ -542,13 +653,11 @@ static int fake_vlr_tx_cm_serv_acc(void *msc_conn_ref)
 	return 0;
 }
 
-static int fake_vlr_tx_cm_serv_rej(void *msc_conn_ref,
-				   enum vlr_proc_arq_result result)
+static int fake_vlr_tx_cm_serv_rej(void *msc_conn_ref, enum gsm48_reject_value cause)
 {
 	struct gsm_subscriber_connection *conn = msc_conn_ref;
-	btw("sending CM Service Reject for %s, result %s",
-	    vlr_subscr_name(conn->vsub),
-	    vlr_proc_arq_result_name(result));
+	btw("sending CM Service Reject for %s, cause: %s",
+	    vlr_subscr_name(conn->vsub), gsm48_reject_value_name(cause));
 	cm_service_result_sent |= RES_REJECT;
 	return 0;
 }
@@ -605,37 +714,77 @@ static int fake_vlr_tx_auth_rej(void *msc_conn_ref)
 	return 0;
 }
 
-static int fake_vlr_tx_ciph_mode_cmd(void *msc_conn_ref, enum vlr_ciph ciph,
-				     bool retrieve_imeisv)
+/* override, requires '-Wl,--wrap=a_iface_tx_cipher_mode' */
+int __real_a_iface_tx_cipher_mode(const struct gsm_subscriber_connection *conn,
+				  struct gsm0808_encrypt_info *ei, int include_imeisv);
+int __wrap_a_iface_tx_cipher_mode(const struct gsm_subscriber_connection *conn,
+				  struct gsm0808_encrypt_info *ei, int include_imeisv)
 {
-	/* FIXME: we actually would like to see the message bytes checked here,
-	 * not possible while msc_vlr_set_ciph_mode() calls
-	 * gsm0808_cipher_mode() directly. When the MSCSPLIT is ready, check
-	 * the tx bytes in the sense of dtap_expect_tx() above. */
-	struct gsm_subscriber_connection *conn = msc_conn_ref;
-	switch (conn->via_ran) {
-	case RAN_GERAN_A:
-		btw("sending Ciphering Mode Command for %s: cipher=%s kc=%s"
-		    " retrieve_imeisv=%d",
-		    vlr_subscr_name(conn->vsub),
-		    vlr_ciph_name(conn->network->a5_encryption),
-		    osmo_hexdump_nospc(conn->vsub->last_tuple->vec.kc, 8),
-		    retrieve_imeisv);
-		break;
-	case RAN_UTRAN_IU:
-		btw("sending SecurityModeControl for %s",
-		    vlr_subscr_name(conn->vsub));
-		break;
-	default:
-		btw("UNKNOWN RAN TYPE %d", conn->via_ran);
-		OSMO_ASSERT(false);
-		return -1;
-	}
+	int i;
+	btw("sending Ciphering Mode Command for %s: include_imeisv=%d",
+	    vlr_subscr_name(conn->vsub), include_imeisv);
+	for (i = 0; i < ei->perm_algo_len; i++)
+		btw("...perm algo: %u", ei->perm_algo[i]);
+	OSMO_ASSERT(ei->key_len <= sizeof(ei->key));
+	btw("...key: %s", osmo_hexdump_nospc(ei->key, ei->key_len));
 	cipher_mode_cmd_sent = true;
-	cipher_mode_cmd_sent_with_imeisv = retrieve_imeisv;
+	cipher_mode_cmd_sent_with_imeisv = include_imeisv;
+
+	if (!cipher_mode_expect_kc
+	    || strcmp(cipher_mode_expect_kc, osmo_hexdump_nospc(ei->key, ei->key_len))) {
+		log("FAILURE: expected kc=%s", cipher_mode_expect_kc ? : "NULL");
+		OSMO_ASSERT(false);
+	}
 	return 0;
 }
 
+/* override, requires '-Wl,--wrap=ranap_iu_tx_sec_mode_cmd' */
+int __real_ranap_iu_tx_sec_mode_cmd(struct ranap_ue_conn_ctx *uectx, struct osmo_auth_vector *vec,
+				    int send_ck, int new_key);
+int __wrap_ranap_iu_tx_sec_mode_cmd(struct ranap_ue_conn_ctx *uectx, struct osmo_auth_vector *vec,
+				    int send_ck, int new_key)
+{
+	btw("sending SecurityModeControl for UE ctx %u send_ck=%d new_key=%d",
+	    uectx->conn_id, send_ck, new_key);
+	btw("...ik=%s", osmo_hexdump_nospc(vec->ik, sizeof(vec->ik)));
+	if (send_ck)
+		btw("...ck=%s", osmo_hexdump_nospc(vec->ck, sizeof(vec->ck)));
+	security_mode_ctrl_sent = true;
+	if (!security_mode_expect_ik
+	    || strcmp(security_mode_expect_ik, osmo_hexdump_nospc(vec->ik, sizeof(vec->ik)))) {
+		log("FAILURE: expected ik=%s", security_mode_expect_ik ? : "NULL");
+		OSMO_ASSERT(false);
+	}
+	if (((!!send_ck) != (!!security_mode_expect_ck))
+	    || (security_mode_expect_ck
+		&& strcmp(security_mode_expect_ck, osmo_hexdump_nospc(vec->ck, sizeof(vec->ck))))) {
+		log("FAILURE: expected ck=%s", security_mode_expect_ck ? : "NULL");
+		OSMO_ASSERT(false);
+	}
+	return 0;
+}
+
+extern int msc_vlr_set_ciph_mode(void *msc_conn_ref, bool umts_aka, bool retrieve_imeisv);
+
+static int fake_vlr_tx_ciph_mode_cmd(void *msc_conn_ref, bool umts_aka, bool retrieve_imeisv)
+{
+	int rc;
+#ifndef BUILD_IU
+	/* If we built without support for IU, fake the IU part here. The root cause is that we don't
+	 * have differing sets of expected outputs for --enable-iu and --disable-iu. */
+	struct gsm_subscriber_connection *conn = msc_conn_ref;
+	if (conn->via_ran == RAN_UTRAN_IU) {
+		DEBUGP(DMM, "-> SECURITY MODE CONTROL %s\n", vlr_subscr_name(conn->vsub));
+		rc = __wrap_ranap_iu_tx_sec_mode_cmd(conn->iu.ue_ctx, &conn->vsub->last_tuple->vec,
+						     0, 1);
+	} else
+#endif
+	rc = msc_vlr_set_ciph_mode(msc_conn_ref, umts_aka, retrieve_imeisv);
+	if (rc)
+		btw("ERROR sending ciphering mode command: rc=%d", rc);
+	return rc;
+}
+
 void ms_sends_security_mode_complete()
 {
 	OSMO_ASSERT(g_conn);
@@ -644,6 +793,22 @@ void ms_sends_security_mode_complete()
 	msc_rx_sec_mode_compl(g_conn);
 }
 
+void bss_sends_clear_complete()
+{
+	btw("BSS sends BSSMAP Clear Complete");
+	OSMO_ASSERT(g_conn);
+	OSMO_ASSERT(g_conn->via_ran == RAN_GERAN_A);
+	msc_subscr_conn_rx_bssmap_clear_complete(g_conn);
+}
+
+void rnc_sends_release_complete()
+{
+	btw("RNC sends Iu Release Complete");
+	OSMO_ASSERT(g_conn);
+	OSMO_ASSERT(g_conn->via_ran == RAN_UTRAN_IU);
+	msc_subscr_conn_rx_iu_release_complete(g_conn);
+}
+
 const struct timeval fake_time_start_time = { 123, 456 };
 
 void fake_time_start()
@@ -653,13 +818,30 @@ void fake_time_start()
 	fake_time_passes(0, 0);
 }
 
-void check_talloc(void *msgb_ctx, void *tall_bsc_ctx, int expected_blocks)
+static void check_talloc(void *msgb_ctx, void *msc_vlr_tests_ctx)
 {
+	/* Verifying that the msgb context is empty */
 	talloc_report_full(msgb_ctx, stderr);
+	/* Expecting these to stick around in msc_vlr_tests_ctx:
+	 * talloc_total_blocks(tall_bsc_ctx) == 12
+	 * full talloc report on 'msc_vlr_tests_ctx' (total   3636 bytes in  12 blocks)
+	 *     struct gsup_client             contains    248 bytes in   1 blocks (ref 0) 0x563a489c05f0
+	 *     struct gsm_network             contains   2031 bytes in   4 blocks (ref 0) 0x563a489bfbb0
+	 *         struct vlr_instance            contains    168 bytes in   1 blocks (ref 0) 0x563a489c04e0
+	 *         no_gsup_server                 contains     15 bytes in   1 blocks (ref 0) 0x563a489c0460
+	 *         ../../../src/libosmocore/src/rate_ctr.c:228 contains   1552 bytes in   1 blocks (ref 0) 0x563a489bfd40
+	 *     logging                        contains   1357 bytes in   5 blocks (ref 0) 0x563a489bf440
+	 *         struct log_target              contains    228 bytes in   2 blocks (ref 0) 0x563a489bf9f0
+	 *             struct log_category            contains     68 bytes in   1 blocks (ref 0) 0x563a489bfb00
+	 *         struct log_info                contains   1128 bytes in   2 blocks (ref 0) 0x563a489bf4b0
+	 *             struct log_info_cat            contains   1088 bytes in   1 blocks (ref 0) 0x563a489bf540
+	 *     msgb                           contains      0 bytes in   1 blocks (ref 0) 0x563a489bf3d0
+	 * (That's 12 counting the root ctx)
+	 */
 	fprintf(stderr, "talloc_total_blocks(tall_bsc_ctx) == %zu\n",
-		talloc_total_blocks(tall_bsc_ctx));
-	if (talloc_total_blocks(tall_bsc_ctx) != expected_blocks)
-		talloc_report_full(tall_bsc_ctx, stderr);
+		talloc_total_blocks(msc_vlr_tests_ctx));
+	if (talloc_total_blocks(msc_vlr_tests_ctx) != 12)
+		talloc_report_full(msc_vlr_tests_ctx, stderr);
 	fprintf(stderr, "\n");
 }
 
@@ -716,11 +898,14 @@ static void handle_options(int argc, char **argv)
 
 void *msgb_ctx = NULL;
 
-void run_tests(int nr)
+static void run_tests(int nr)
 {
 	int test_nr;
-	nr --; /* arg's first test is 1, in here it's 0 */
-	for (test_nr = 0; msc_vlr_tests[test_nr]; test_nr ++) {
+
+	check_talloc(msgb_ctx, msc_vlr_tests_ctx);
+
+	nr--; /* arg's first test is 1, in here it's 0 */
+	for (test_nr = 0; msc_vlr_tests[test_nr]; test_nr++) {
 		if (nr >= 0 && test_nr != nr)
 			continue;
 
@@ -732,36 +917,21 @@ void run_tests(int nr)
 		if (cmdline_opts.verbose)
 			fprintf(stderr, "(test nr %d)\n", test_nr + 1);
 
-		check_talloc(msgb_ctx, tall_bsc_ctx, 9);
-	} while(0);
+		check_talloc(msgb_ctx, msc_vlr_tests_ctx);
+	}
 }
 
-int main(int argc, char **argv)
+struct gsm_network *test_net(void *ctx)
 {
-	handle_options(argc, argv);
+	struct gsm_network *net = gsm_network_init(ctx, mncc_recv);
 
-	tall_bsc_ctx = talloc_named_const(NULL, 0, "subscr_conn_test_ctx");
-	msgb_ctx = msgb_talloc_ctx_init(tall_bsc_ctx, 0);
-	osmo_init_logging(&info);
-
-	_log_lines = cmdline_opts.verbose;
-
-	OSMO_ASSERT(osmo_stderr_target);
-	log_set_use_color(osmo_stderr_target, 0);
-	log_set_print_timestamp(osmo_stderr_target, 0);
-	log_set_print_filename(osmo_stderr_target, _log_lines? 1 : 0);
-	log_set_print_category(osmo_stderr_target, 1);
-
-	net = gsm_network_init(tall_bsc_ctx, 1, 1, fake_mncc_recv);
 	net->gsup_server_addr_str = talloc_strdup(net, "no_gsup_server");
 	net->gsup_server_port = 0;
 
-	osmo_fsm_log_addr(false);
 	OSMO_ASSERT(msc_vlr_alloc(net) == 0);
 	OSMO_ASSERT(msc_vlr_start(net) == 0);
 	OSMO_ASSERT(net->vlr);
 	OSMO_ASSERT(net->vlr->gsup_client);
-	msc_subscr_conn_init();
 
 	net->vlr->ops.tx_lu_acc = fake_vlr_tx_lu_acc;
 	net->vlr->ops.tx_lu_rej = fake_vlr_tx_lu_rej;
@@ -771,6 +941,34 @@ int main(int argc, char **argv)
 	net->vlr->ops.tx_auth_rej = fake_vlr_tx_auth_rej;
 	net->vlr->ops.set_ciph_mode = fake_vlr_tx_ciph_mode_cmd;
 
+	return net;
+}
+
+int main(int argc, char **argv)
+{
+	handle_options(argc, argv);
+
+	msc_vlr_tests_ctx = talloc_named_const(NULL, 0, "msc_vlr_tests_ctx");
+	msgb_ctx = msgb_talloc_ctx_init(msc_vlr_tests_ctx, 0);
+	osmo_init_logging2(msc_vlr_tests_ctx, &info);
+
+	_log_lines = cmdline_opts.verbose;
+
+	OSMO_ASSERT(osmo_stderr_target);
+	log_set_use_color(osmo_stderr_target, 0);
+	log_set_print_timestamp(osmo_stderr_target, 0);
+	log_set_print_filename(osmo_stderr_target, _log_lines? 1 : 0);
+	log_set_print_category(osmo_stderr_target, 1);
+
+	if (cmdline_opts.verbose)
+		log_set_category_filter(osmo_stderr_target, DLSMS, 1, LOGL_DEBUG);
+
+	net = test_net(msc_vlr_tests_ctx);
+
+	osmo_fsm_log_addr(false);
+
+	msc_subscr_conn_init();
+
 	clear_vlr();
 
 	if (optind >= argc)
@@ -779,6 +977,7 @@ int main(int argc, char **argv)
 		int arg;
 		long int nr;
 		for (arg = optind; arg < argc; arg++) {
+			errno = 0;
 			nr = strtol(argv[arg], NULL, 10);
 			if (errno) {
 				fprintf(stderr, "Invalid argument: %s\n",
@@ -792,8 +991,6 @@ int main(int argc, char **argv)
 
 	printf("Done\n");
 
-	talloc_free(the_bts);
-
-	check_talloc(msgb_ctx, tall_bsc_ctx, 9);
+	check_talloc(msgb_ctx, msc_vlr_tests_ctx);
 	return 0;
 }
